#include "shm.h"

/*
 * shmcreate() - create the given amount of shared memory and return the 
 *	pointer to this memory.  Return NULL on failure.  We have imposed
 *	a limit of 1024 bytes for the total amount of shared memory created.
 *      Attaches if the key exists.
 */
char *shmcreate(key_t key, int cnt /* number of bytes */) {
  int i;

  if (cShm >= MAXSHM)
    return(NULL);		/* no more room to remember */

  if (sbShmBase == NULL) {
    /* need to get a shared memory chunk */
    if (cnt > MAXMEM)
      return(NULL);		/* will never get this much room */
    
    if ((shm = shmget(key, MAXMEM, 0666|IPC_CREAT|IPC_EXCL)) < 0) {
      if (errno != EEXIST)
	return(NULL);
      else {
	/* There is already a segment there.  Attach to it. */
	if ((shm = shmget(key, MAXMEM, 0666|IPC_EXCL)) < 0)
	   return(NULL);
      }
    }
    if ((int)(sbShmBase = (char *)shmat(shm, NULL, 0)) == -1) {
      /* a problem */
      (void)shmctl(shm, IPC_RMID, (struct shmid_ds *)NULL);
      return(NULL);
    }

  }
  
  /* ok if we get to this point then we have a shared memory chunk */
  if ((cByteOffset + cnt) > MAXMEM)
    return(NULL);	/* more memory than we have */
  else {
    /* remember this offset */
    for (i = 0; i < cShm; i++)
      if (rgsbShm[i] == NULL)
	break;
    rgsbShm[i] = &sbShmBase[cByteOffset];
    cShm++;			/* one more mapping */
    if (cnt%8 == 0)
      cByteOffset += cnt;	/* just update offset by the number of bytes */
    else
      cByteOffset += cnt + 8 - cnt%8; /* make divisible by 8 */

    return(rgsbShm[i]);
  }
}

/*
 * shmdelete() - delete the shared memory associated with the given address.
 *	Return -1 on failure, zero on success.
 */
int shmdelete(char *sbShm) {
  int i, ret;
  
  for (i = 0; i < MAXSHM; i++) {
    if (rgsbShm[i] == sbShm) {
      /* note we do not try and do compaction of the space! */
      ret = 0;
      if (cShm <= 1) {
	/* have removed the last piece, so remove entire chunk */
	(void)shmdt(sbShmBase);
	ret = shmctl(shm, IPC_RMID, (struct shmid_ds *)NULL);
	if (ret == 0) {
	  sbShmBase = NULL;
	  shm = -1;
	}
      }
      if (ret == 0) {
	cShm--;
	rgsbShm[i] = NULL; /* free up space */
      }
      return(ret);
    }
  }
  return(-1);			/* could not find this memory piece */
}

/*
 * semwaitn() - wait on the given semaphore for the given count.  Return -1 on 
 *	error, zero on success.
 */
int semwaitn(int sem, int cnt) {
    struct sembuf sembuf;

    if (cnt <= 0)
	return(0);		/* return immediately */
    sembuf.sem_num = sem;
    sembuf.sem_op = -cnt;
    sembuf.sem_flg = 0;
    while (semop(semBase, &sembuf, 1) == -1) {
	if (errno != EINTR)
	    return(-1);		/* do not return on an interrupted wait */
    }
    return(0);
}

/*
 * semwait() - wait on the given semaphore.  Return -1 on error, zero on
 *	success.
 */
int semwait(int sem) {
    struct sembuf sembuf;

#ifdef DEBUG
    printf("In semwait(). semval: %d\n", semctl(semBase, sem, GETVAL));
#endif

    sembuf.sem_num = sem;
    sembuf.sem_op = -1;
    sembuf.sem_flg = 0;
    while (semop(semBase, &sembuf, 1) == -1) {
	if (errno != EINTR)
	    return(-1);		/* do not return on an interrupted wait */
    }
    return(0);
}

/*
 * semsignal() - signal the given semaphore.  Returns -1 on failure.
 */
int semsignal(int sem) {
    struct sembuf sembuf;

#ifdef DEBUG
    printf("In semsignal(). semval: %d\n", semctl(semBase, sem, GETVAL));
#endif

    sembuf.sem_num = sem;
    sembuf.sem_op = 1;
    sembuf.sem_flg = 0;
    return(semop(semBase, &sembuf, 1));
}

/*
 * sgetval() - get the semaphore integer value.
 */
int sgetval(int sem) {
#ifdef LINUX
    return semctl(semBase, sem, GETVAL, NULL);
#else
    return semctl(semBase, sem, GETVAL);
#endif
}


/*
 * semsignaln() - signal the given semaphore, the given number of times.  Returns
 *	-1 on failure.
 */
int semsignaln(int sem, int cnt) {
    struct sembuf sembuf;

    if (cnt <= 0)
	return(0);		/* return immediately */
    sembuf.sem_num = sem;
    sembuf.sem_op = cnt;
    sembuf.sem_flg = 0;
    return(semop(semBase, &sembuf, 1));
}

/*
 * screate() - create a semaphore with the given initial count and the key.  
 *      Returns -1 on failure, a semaphore id on success.  
 *      We have imposed a limit 12 semaphores that can be created.
 */
int screate(key_t key, int cnt) {
    union semun arg;

    if (iSem >= MAXSEM)
	return(-1);		/* have reached semaphore limit */

#ifdef DEBUG
    printf("key is %d\n", key);
    printf("sembase is %d\n", semBase);
#endif

    if (semBase == -1) {
	/* need to get a chunk of semaphores */
      if ((semBase = semget(key, MAXSEM, 0666)) < 0) {
#ifdef DEBUG
	printf("Could not attach to the semaphore.\n");
#endif
      } else {
#ifdef DEBUG
	  printf("Semaphore exists.  Attaching to it. \n");
#endif
	  rgSem[iSem] = iSem;		/* put index value into array */
	  cSem++;
	  iSem++;
	  return(iSem-1);
      }
#ifdef DEBUG
      printf("Creating new semaphore.\n");
#endif
      if ((semBase = semget(key, MAXSEM, 0666|IPC_CREAT|IPC_EXCL)) < 0) {
	 return(-1);
      }
    }
    
    /* ok, at this point we have a set of semaphores */
    arg.val = cnt;
    if (semctl(semBase, iSem, SETVAL, arg) < 0) {
      if (cSem == 0) {
	semctl(semBase, 0, IPC_RMID, (union semun *)NULL);
      }
      return(-1);
    }
    rgSem[iSem] = iSem;		/* put index value into array */
    cSem++;
    iSem++;
    return(iSem-1);
}

/*
 * sdelete() - delete the semaphore corresponding to the given sem id.  Returns
 *	-1 on failure, zero on success. 
 */
int sdelete(int sem) {
    int ret;

    /* note we are not trying to reuse semaphore slots so we do nothing with
       iSem */
    if (rgSem[sem] != sem)
	return(-1);		/* does not match up with what it should */

    ret = 0;
    if (cSem <= 1) {
	ret = semctl(semBase, 0, IPC_RMID, (union semun *)NULL);
	if (ret == 0)
	    semBase = -1;
    }
    if (ret == 0) {
	cSem--;
	rgSem[sem] = -1;		/* no longer in use */
    }
    return(ret);
}

/*
 * scount() - return the count of the given sem id.  -1 on failure.
 */
int scount(int sem) {
    return(semctl(semBase, sem, GETVAL, (union semun *)NULL));
}

/*
 * sreset() - reset the count of the given sem id.  Return -1 on failure,
 *	zero on success.
 */
int sreset(int sem, int cnt) {
    union semun arg;

    /* need to free up processes in the queue, before reseting the count.
       need to check manual page to see if this action really happens,
       otherwise we need to do it. */
    arg.val = cnt;
    return(semctl(semBase, sem, SETVAL, arg));
}

